<?php

namespace Drupal\consistent_breadcrumbs;

use Drupal\Core\Breadcrumb\Breadcrumb;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;

class ConsistentBreadcrumbManager implements ConsistentBreadcrumbManagerInterface {

  use StringTranslationTrait;

  /**
   * The static cache.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  private $staticCache;

  /**
   * The breadcrumb manager.
   *
   * @var \Drupal\Core\Breadcrumb\BreadcrumbBuilderInterface
   */
  private $breadcrumbManager;

  /**
   * The routing helper.
   *
   * @var \Drupal\consistent_breadcrumbs\ConsistentBreadcrumbsRoutingHelper
   */
  private $routingHelper;

  /**
   * The consistent breadcrumb builders, keyed by priority.
   *
   * @var \Drupal\consistent_breadcrumbs\ConsistentBreadcrumbBuilderInterface[]
   */
  private $builders = [];

  /**
   * The sorted consistent breadcrumb builders.
   *
   * @var \Drupal\consistent_breadcrumbs\ConsistentBreadcrumbBuilderInterface[]
   */
  private $sortedBuilders;

  /**
   * ConsistentBreadcrumbManager constructor.
   *
   * @param \Drupal\Core\Cache\CacheBackendInterface $staticCache
   * @param \Drupal\Core\Breadcrumb\BreadcrumbBuilderInterface $breadcrumbManager
   * @param \Drupal\consistent_breadcrumbs\ConsistentBreadcrumbsRoutingHelper $routingHelper
   */
  public function __construct(\Drupal\Core\Cache\CacheBackendInterface $staticCache, \Drupal\Core\Breadcrumb\BreadcrumbBuilderInterface $breadcrumbManager, \Drupal\consistent_breadcrumbs\ConsistentBreadcrumbsRoutingHelper $routingHelper) {
    $this->staticCache = $staticCache;
    $this->breadcrumbManager = $breadcrumbManager;
    $this->routingHelper = $routingHelper;
  }

  /**
   * {@inheritdoc}
   */
  public function addBuilder(ConsistentBreadcrumbBuilderInterface $builder, $priority) {
    $this->builders[$priority][] = $builder;
    // Force the builders to be re-sorted.
    $this->sortedBuilders = NULL;
  }

  /**
   * Returns the sorted array of breadcrumb builders.
   *
   * @return \Drupal\Core\Breadcrumb\BreadcrumbBuilderInterface[]
   *   An array of breadcrumb builder objects.
   */
  protected function getSortedBuilders() {
    if (!isset($this->sortedBuilders)) {
      // Sort the builders according to priority.
      krsort($this->builders);
      // Merge nested builders from $this->builders into $this->sortedBuilders.
      $this->sortedBuilders = [];
      foreach ($this->builders as $builders) {
        $this->sortedBuilders = array_merge($this->sortedBuilders, $builders);
      }
    }
    return $this->sortedBuilders;
  }

  /**
   * @inheritDoc
   */
  public function applies(RouteMatchInterface $route_match) {
    return (bool) $this->cachedMaybeBuild($route_match);
  }

  /**
   * @inheritDoc
   */
  public function build(RouteMatchInterface $routeMatch) {
    $breadcrumb = $this->cachedMaybeBuild($routeMatch);
    assert($breadcrumb);
    return $breadcrumb;
  }

  /**
   * Build the breadcrumb, cached.
   *
   * To compute ::applies(), we have to do all the work, wo we spare re-doing
   * everything on ::build().
   *
   * @param \Drupal\Core\Routing\RouteMatchInterface $routeMatch
   *
   * @return \Drupal\Core\Breadcrumb\Breadcrumb|null
   */
  private function cachedMaybeBuild(RouteMatchInterface $routeMatch) {
    $cid = serialize([$routeMatch->getRouteName(), $routeMatch->getRawParameters()]);
    if ($hit = $this->staticCache->get($cid)) {
      $breadcrumb = $hit->data;
    }
    else {
      $breadcrumb = $this->maybeBuild($routeMatch);
      $this->staticCache->set($cid, $breadcrumb);
    }
    return $breadcrumb;
  }

  /**
   * Build the breadcrumbs if possible.
   *
   * @param \Drupal\Core\Routing\RouteMatchInterface $routeMatch
   *
   * @return \Drupal\Core\Breadcrumb\Breadcrumb
   */
  private function maybeBuild(RouteMatchInterface $routeMatch) {
    $breadcrumb = new Breadcrumb();
    $cacheability = new CacheableMetadata();
    foreach ($this->getSortedBuilders() as $builder) {
      $breadcrumbSlice = $builder->getBreadcrumbSlice($routeMatch, $isFinalLink);
      // If this breadcrumb builder does not match, the final result still
      // depends on the non-match, so add it here.
      $cacheability->addCacheableDependency($breadcrumbSlice);
      $links = $breadcrumbSlice->getLinks();
      if ($links) {
        foreach ($links as $link) {
          $linkRouteMatch = $this->routingHelper->routeMatchFromUrl($link->getUrl(), $accessCacheability, $routeTitle);
          // Do not pursue this breadcrumb slice if any of the links does not
          // have access. Also no need to collect cacheability further, as the
          // final result does not depend on that. Nevertheless keep
          // cacheability of this non-access, as any breadcrumb trail from a
          // later builder depends on the earlier failing access checks.
          $cacheability->addCacheableDependency($accessCacheability);
          if (!$linkRouteMatch) {
            continue 2;
          }
          // Set link text to route title if empty.
          if (!$link->getText()) {
            $link->setText($routeTitle);
          }
        }
        $lastLink = end($links);
        $lastRouteMatch = $this->routingHelper->routeMatchFromUrl($lastLink->getUrl());

        // Default to final link is frontpage.
        if (is_null($isFinalLink)) {
          $isFinalLink = $this->routingHelper->urlIsFront($lastLink->getUrl());
        }

        if ($isFinalLink) {
          $finalBreadcrumbSlice = new Breadcrumb();
        }
        else {
          $consistentBreadcrumbs = $this->maybeBuild($lastRouteMatch);
          $cacheability->addCacheableDependency($consistentBreadcrumbs);
          if ($consistentBreadcrumbs->getLinks()) {
            $finalBreadcrumbSlice = $consistentBreadcrumbs;
          }
          // Note that this discards non matching crumb cacheability.
          elseif ($this->breadcrumbManager->applies($lastRouteMatch)) {
            $finalBreadcrumbSlice = $this->breadcrumbManager->build($lastRouteMatch);
          }
        }

        if (isset($finalBreadcrumbSlice)) {
          $breadcrumb = $this->mergeBreadcrumbs($breadcrumb, $finalBreadcrumbSlice, $breadcrumbSlice);
          break;
        }
      }
    }
    // Add cacheability of non matching builders and failed access checks.
    $breadcrumb->addCacheableDependency($cacheability);
    return $breadcrumb;
  }

  private function mergeBreadcrumbs(Breadcrumb ...$breadcrumbs) {
    $merged = new Breadcrumb();
    foreach ($breadcrumbs as $breadcrumb) {
      // Append the links and transfer cacheability.
      foreach ($breadcrumb->getLinks() as $link) {
        $merged->addLink($link);
      }
      $merged->addCacheableDependency($breadcrumb);
    }
    return $merged;
  }

}
