<?php

namespace Drupal\consistent_breadcrumbs;

use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Cache\RefinableCacheableDependencyInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Controller\TitleResolverInterface;
use Drupal\Core\Routing\AccessAwareRouterInterface;
use Drupal\Core\Routing\RouteMatch;
use Drupal\Core\Url;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Exception\AccessDeniedHttpException;

class ConsistentBreadcrumbsRoutingHelper {

  /**
   * @var \Drupal\Core\Routing\AccessAwareRouterInterface
   */
  private $accessAwareRouter;

  /**
   * The title resolver.
   *
   * @var \Drupal\Core\Controller\TitleResolverInterface
   */
  private $titleResolver;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  private $configFactory;

  /**
   * Static cache for ::urlIsFront
   *
   * @var bool[]
   */
  private $urlIsFront;

  /**
   * Static cache for frontPagePath().
   *
   * @var string;
   */
  private $frontPagePath;

  /**
   * ConsistentBreadcrumbsRoutingHelper constructor.
   *
   * @param \Drupal\Core\Routing\AccessAwareRouterInterface $accessAwareRouter
   * @param \Drupal\Core\Controller\TitleResolverInterface $titleResolver
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   */
  public function __construct(AccessAwareRouterInterface $accessAwareRouter, TitleResolverInterface $titleResolver, ConfigFactoryInterface $configFactory) {
    $this->accessAwareRouter = $accessAwareRouter;
    $this->titleResolver = $titleResolver;
    $this->configFactory = $configFactory;
  }


  /**
   * Create a RouteMatch from an Url object.
   *
   * @param \Drupal\Core\Url $url
   *   The URL.
   * @param \Drupal\Core\Cache\RefinableCacheableDependencyInterface|null $accessCacheability
   *   The access cacheability returned.
   * @param null $routeTitle
   *   The route title, returned by reference.
   *
   * @return \Drupal\Core\Routing\RouteMatchInterface|null
   *   The route match, or NULL if no route or current user has no access.
   */
  public function routeMatchFromUrl(Url $url, RefinableCacheableDependencyInterface &$accessCacheability = NULL, &$routeTitle = NULL) {
    $accessCacheability = new CacheableMetadata();
    $request = Request::create($url->toString());
    try {
      // We can't use RouteMatch::createFromRequest without first having the
      // router populate the request attributes.
      $this->accessAwareRouter->matchRequest($request);
    } catch (AccessDeniedHttpException $e) {
      return NULL;
    }
    // Add access cacheability.
    // @see \Drupal\Core\Routing\AccessAwareRouter::checkAccess
    assert($request->attributes->has(AccessAwareRouterInterface::ACCESS_RESULT));
    $accessResult = $request->attributes->get(AccessAwareRouterInterface::ACCESS_RESULT);
    $accessCacheability->addCacheableDependency($accessResult);
    // Retrieve route title.
    $routeMatch = RouteMatch::createFromRequest($request);
    $routeTitle = $this->titleResolver->getTitle($request, $routeMatch->getRouteObject());
    return $routeMatch;
  }

  /**
   * Check if an Url is identical to the front page.
   *
   * Checking this in not trivial as the front page may have aliases.
   *
   * @param \Drupal\Core\Url $url
   *   The url.
   *
   * @return bool
   *   True if the the url is identical to the front page.
   *
   * @see \Drupal\Core\Path\PathMatcher::isFrontPage
   */
  public function urlIsFront(Url $url) {
    $cid = $url->toUriString();
    if (!isset($this->urlIsFront[$cid])) {
      $routeMatch = $this->routeMatchFromUrl($url);
      $matchedUrl = Url::fromRouteMatch($routeMatch);
      $this->urlIsFront[$cid] = $matchedUrl->getRouteName()
        && '/' . $matchedUrl->getInternalPath() === $this->getFrontPagePath();
    }
    return $this->urlIsFront[$cid];
  }

  /**
   * Gets the current front page path.
   *
   * @return string
   *   The front page path.
   */
  protected function getFrontPagePath() {
    // Lazy-load front page config.
    if (!isset($this->frontPagePath)) {
      $this->frontPagePath = $this->configFactory->get('system.site')
        ->get('page.front');
    }
    return $this->frontPagePath;
  }

}
